from pathlib import Path

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.small_tools import is_pathname_valid
from pkglts.version import __version__


class OptionSphinx(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            theme="default",
            autodoc_dvlpt=True,
            build_dir="build/sphinx",
            gallery="",
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        if cfg['doc']['fmt'] != 'rst':
            invalids.append('doc.fmt')

        theme = cfg[self._name]['theme']
        if theme != str(theme):
            invalids.append('sphinx.theme')

        gallery = cfg[self._name]['gallery']
        if gallery != "" and not is_pathname_valid(gallery):
            invalids.append('sphinx.gallery')

        return invalids

    def require_option(self):
        return ['pysetup']

    def require(self, cfg):
        yield Dependency('sphinx', intent='doc')
        if cfg["sphinx"]["theme"] == "sphinx_rtd_theme":
            yield Dependency('sphinx_rtd_theme', intent='doc')

        if cfg['sphinx']['gallery'] != "":
            yield Dependency('sphinx-gallery', intent='doc')



import pytest

from pkglts.config_management import Config
from pkglts.local import (pkg_full_name, src_dir)


def test_pkg_full_name():
    cfg = {}
    with pytest.raises(KeyError):
        pkg_full_name(Config(cfg))

    cfg['base'] = dict(pkgname='toto', namespace=None)
    name1 = pkg_full_name(Config(cfg))
    cfg['base']['namespace'] = 'oa'
    name2 = pkg_full_name(Config(cfg))

    assert name1 != name2


def test_src_dir():
    cfg = {}
    with pytest.raises(KeyError):
        pkg_full_name(Config(cfg))

    cfg['base'] = dict(pkgname='toto', namespace=None)
    dir1 = src_dir(Config(cfg))
    cfg['base']['namespace'] = 'oa'
    dir2 = src_dir(Config(cfg))

    assert dir1 != dir2



from pathlib import Path

from pkglts.dependency import Dependency
from pkglts.option.doc import fmt_badge
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionCoveralls(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self):
        return ['coverage', 'travis']

    def require(self, cfg):
        del cfg
        yield Dependency('coveralls', intent='test')

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project']

        url = f"coveralls.io/github/{owner}/{project}?branch=master"
        img = f"coveralls.io/repos/github/{owner}/{project}/badge.svg?branch=master"
        badge = fmt_badge(img, url, "Coverage report status", cfg['doc']['fmt'])

        return {"badge": badge}



from os import remove
from os.path import exists

import pytest
from pkglts.file_management import write_file


@pytest.fixture()
def tmp_pth():
    pth = 'toto.txt'

    yield pth

    if exists(pth):
        remove(pth)


def test_write_file(tmp_pth):
    write_file(tmp_pth, "lorem ipsum")
    assert exists(tmp_pth)



from pkglts.config_management import Config


def test_requirements():
    cfg = Config(dict(test={'suite_name': 'pytest'},
                      reqs={'require': []}))
    cfg.load_extra()
    assert len(cfg._env.globals['reqs'].requirements('install')) == 0
    assert len(cfg._env.globals['reqs'].requirements('test')) == 2
    assert len(cfg._env.globals['reqs'].intents()) >= 4
    assert len(cfg._env.globals['reqs'].conda_reqs('install')) >= 0
    assert len(cfg._env.globals['reqs'].pip_reqs('install')) >= 0



import logging
import re
import subprocess
from pathlib import Path

from pkglts.option_object import Option
from pkglts.version import __version__
from unidecode import unidecode

LOGGER = logging.getLogger(__name__)


class OptionGit(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self):
        return ['base']

    def environment_extensions(self, cfg):
        del cfg

        try:
            log = subprocess.check_output(['git', 'log', '--all', '--use-mailmap'],
                                          stderr=subprocess.STDOUT).decode('utf-8')
            commiters = re.findall(r'Author: (.* <.*@.*>)\n', unidecode(log))
            ccs = [(commiters.count(name), name) for name in set(commiters)]
            contributors = [name for nb, name in sorted(ccs, reverse=True)]
        except (KeyError, OSError):
            LOGGER.warning("Please add git to your $PATH")
            contributors = ["I failed to construct the contributor list"]
        except subprocess.CalledProcessError as err:
            contributors = [f"Pb with git, {err}"]

        return {'contributors': contributors}



from pathlib import Path

from pkglts.option.doc import fmt_badge
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionRequires(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self):
        return ['doc', 'github']

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project']

        base_url = f"requires.io/github/{owner}/{project}/"
        url = f"{base_url}requirements/?branch=master"
        img = f"{base_url}requirements.svg?branch=master"
        badge = fmt_badge(img, url, "Requirements status", cfg['doc']['fmt'])

        return {"badge": badge}



""" Specific helper function for manage script
"""

import logging

from .hash_management import compute_hash, pth_as_key
from .local import init_namespace_dir
from .option_tools import available_options, get_user_permission
from .templating import Template

LOGGER = logging.getLogger(__name__)

TPL_SRC_NAME = "{" + "{ base.pkgname }" + "}"


def check_option_parameters(name, cfg):
    """Check that the parameters associated to an option are valid.

    Args:
        name (str): option name
        cfg (Config):  current package configuration
    """
    opt = available_options[name]
    return opt.check(cfg)


def update_opt(name, cfg):
    """Update an option of this package.

    Notes: If the option does not exists yet, add it first.
           See the list of available option online

    Args:
        name (str): name of option to add
        cfg (Config):  current package configuration
    """
    LOGGER.info("update option %s", name)

    # test existence of option
    try:
        opt = available_options[name]
    except KeyError:
        raise KeyError(f"option '{name}' does not exists")

    # find other option requirements in repository
    for option_name in opt.require_option():
        if option_name not in cfg.installed_options():
            LOGGER.info("need to install option '%s' first", option_name)
            if (cfg["_pkglts"]['auto_install'] or
                    get_user_permission("install")):
                cfg = update_opt(option_name, cfg)
            else:
                return cfg

    # find parameters required by option config
    opt.update_parameters(cfg.template())
    cfg.resolve()

    return cfg


def find_templates(src_dir, tgt_dir, cfg, rg_tree):
    """Find all template files associated to an option

    Warnings: modify rg_tree in place

    Args:
        src_dir (Path): path to reference files
        tgt_dir (Path): path to target where files will be written
        cfg (Config):  current package configuration
        rg_tree (dict): Structure to store path to templates found

    Returns:
        None
    """
    LOGGER.debug("find_templates in %s", src_dir)

    for src_pth in src_dir.glob("*"):
        tgt_name = cfg.render(src_pth.name)
        if tgt_name.endswith(".tpl"):
            tgt_name = tgt_name[:-4]

        tgt_pth = tgt_dir / tgt_name
        # handle namespace
        if src_pth.is_dir() and src_dir.name == 'src' and src_pth.name == TPL_SRC_NAME:
            namespace = cfg['base']['namespace']
            if namespace is not None:
                ns_pth = tgt_dir / namespace
                init_namespace_dir(ns_pth, rg_tree)
                tgt_pth = ns_pth / tgt_name

        if src_pth.is_dir():
            if tgt_name in ("", "_"):
                # do nothing
                pass
            else:
                find_templates(src_pth, tgt_pth, cfg, rg_tree)
        else:
            if tgt_name.split('.')[0] == "_":
                pass
            else:
                try:
                    rg_tree[pth_as_key(tgt_pth)].append(src_pth)
                except KeyError:
                    rg_tree[pth_as_key(tgt_pth)] = [src_pth]


def render_template(src_pths, tgt_pth, cfg, overwrite_file):
    """Render template and write it in tgt_pth

    Args:
        src_pths (list): list of reference templates
        tgt_pth (Path): path to created file
        cfg (Config):  current package configuration
        overwrite_file (dict): which files to overwrite

    Returns:
        (dict|None): bid, hash of block content
    """
    LOGGER.debug("render file '%s'", tgt_pth)
    if not overwrite_file.get(pth_as_key(tgt_pth), True):
        LOGGER.debug("no overwrite")
        return None

    tpl = Template()
    for src_pth in src_pths:
        tpl.parse(src_pth)

    blocks = tpl.render(cfg, tgt_pth)
    if blocks is None:
        return None
    else:  # non binary file
        return dict((bid, compute_hash(cnt)) for bid, cnt in blocks)



# from pkglts.option.version.handlers import fetch_github_version
#
#
# def test_handlers():
#     pkg_cfg = dict()
#     txt = fetch_github_version("", pkg_cfg)
#     assert txt == ""


# def test_handlers_not_available(mocker):
#     def import_call(*args):
#         print args
#
#     mocker.patch("pkglts.option.version.handlers.__builtins__",
#                     new_callable=import_call)
#     # from pkglts.option.version.handlers import fetch_github_version
#     assert fetch_github_version == "0"



import pytest

from pkglts.config_management import Config, DEFAULT_CFG
from pkglts.manage import add_option


def test_manage_add_opt_raise_error_if_already_installed():
    cfg = Config(DEFAULT_CFG)
    cfg = add_option("base", cfg)
    with pytest.raises(UserWarning):
        add_option('base', cfg)



"""
Set of functions to extend jinja2.
"""
import re

from .small_tools import ensure_path

OPENING_MARKER = "{" + "#"
CLOSING_MARKER = "#" + "}"

BLOCK_RE = re.compile(
    r"\{#[ ]pkglts,[ ](?P<key>[a-zA-Z0-9._]*)(?P<aft_head>.*?)\n(?P<cnt>.*?)#\}(?P<aft_foot>.*?(?=\n))",
    re.DOTALL | re.MULTILINE)

NON_BIN_EXT = ("", ".bat", ".cfg", ".in", ".ini", ".md", ".no", ".ps1", ".py", ".rst", ".sh",
               ".svg", ".tpl", ".txt", ".yml", ".yaml")


class TplBlock:
    """Simple container for template blocks."""

    def __init__(self):
        self.before_header = ""
        self.bid = None
        self.loc = None
        self.after_header = ""
        self.content = ""
        self.before_footer = ""
        self.after_footer = ""

    def __str__(self):
        return (f"Block(\n"
                f"|{self.before_header}|{self.bid}|{self.after_header}|\n"
                f"|{repr(self.content)}|\n|{self.before_footer}|{self.after_footer}|\n)")

    def pkglts_defined(self, bid, cnt):
        """Create a block regenerated by pkglts.

        Args:
            bid (str): id for this block (unique per file only)
            cnt (str): content

        Returns:
            None
        """
        self.bid = bid
        self.content = cnt

    def user_defined(self, cnt):
        """Make this block a user defined one.

        Args:
            cnt (str): content

        Returns:
            None
        """
        self.bid = None
        self.content = cnt


def parse_source(txt):
    """Parse text to find preserved blocks

    Args:
        txt (str): full text to parse

    Returns:
        (list of [str, str, str, str]): ordered list of blocks:
                    - block_id (or None if content is not preserved
                    - line start before start for preserved content
                    - content
                    - line start before end for preserved content
    """
    blocks = []
    last_end = -1

    for res in BLOCK_RE.finditer(txt):
        i = res.start()
        while i > last_end and txt[i] != "\n":
            i -= 1

        if i > last_end:
            block = TplBlock()
            block.user_defined(txt[(last_end + 1): (i + 1)])
            blocks.append(block)

        bef = txt[(i + 1): res.start()]

        bid = res.group('key')
        aft_head = res.group('aft_head')
        cnt = res.group('cnt')
        aft_foot = res.group('aft_foot')

        i = len(cnt) - 1
        while i > 0 and cnt[i] != "\n":
            i -= 1
        aft = cnt[(i + 1): len(cnt)]
        cnt = cnt[:(i + 1)]
        last_end = res.end()

        block = TplBlock()
        block.pkglts_defined(bid, cnt)
        block.before_header = bef
        block.after_header = aft_head
        block.before_footer = aft
        block.after_footer = aft_foot

        if aft_head.startswith(", after"):
            block.loc = ('after', aft_head[7:].split(" ")[1])
        elif aft_head.startswith(", before"):
            block.loc = ('before', aft_head[8:].split(" ")[1])
        elif aft_head.startswith(", replace"):
            block.loc = ('replace', aft_head[8:].split(" ")[1])

        blocks.append(block)

    if last_end < (len(txt) - 1):
        block = TplBlock()
        block.user_defined(txt[(last_end + 1): len(txt)])
        blocks.append(block)

    return blocks


class Template(object):
    """Simple container that holds a list of blocks

    The may objective is to represent a full templated file
    """

    def __init__(self):
        self.blocks = []
        self.bin_cnt = None  # binary content for binary files

    def add(self, block):
        """Insert a new block in the overall template

        Notes: will use block loc to position it.

        Args:
            block (TplBlock): block with all fields filled

        Returns:
            None
        """
        if block.loc is None:
            self.blocks.append(block)
        else:
            ind = [b.bid for b in self.blocks].index(block.loc[1])
            if block.loc[0] == 'after':
                self.blocks.insert(ind + 1, block)
            elif block.loc[0] == 'before':
                self.blocks.insert(ind, block)
            elif block.loc[0] == 'replace':
                self.blocks[ind] = block
            else:
                raise NotImplementedError

    def parse(self, pth):
        """Parse file and append all blocks into this template

        Args:
            pth (Path): path to file to read

        Returns:
            None
        """
        if pth.suffix in NON_BIN_EXT:
            for block in parse_source(pth.read_text()):
                self.add(block)
        else:
            self.bin_cnt = pth.read_bytes()

    def render(self, cfg, tgt_pth):
        """Render template into tgt_pth

        Notes: keeps 'preserved' block structure

        Args:
            cfg (Config):  current package configuration
            tgt_pth (Path): path to potentially non existent yet target file

        Returns:
            (list of [str, str]): key, cnt for preserved blocks
        """
        if self.bin_cnt is not None:
            ensure_path(tgt_pth)
            tgt_pth.write_bytes(self.bin_cnt)
        else:
            return self._render_non_bin(cfg, tgt_pth)

    def _render_non_bin(self, cfg, tgt_pth):
        blocks = []
        if tgt_pth.exists():  # retrieves preserved blocks from source
            # parse tgt file to find 'preserved' blocks
            tgt_blocks = parse_source(tgt_pth.read_text())

            tpl_bids = set(b.bid for b in self.blocks if b.bid is not None)
            tgt_bids = set(b.bid for b in tgt_blocks if b.bid is not None)
            if tpl_bids != tgt_bids:
                raise UserWarning(f"File '{tgt_pth}' not compatible with current template")

            src_blocks = dict((b.bid, b) for b in self.blocks if b.bid is not None)
            for tgt_block in tgt_blocks:
                # reset content of preserved blocks only
                if tgt_block.bid is not None:
                    tgt_block.content = src_blocks[tgt_block.bid].content

                blocks.append(tgt_block)
        else:  # format non preserved blocks for the first and only time
            for block in self.blocks:
                if block.bid is None:
                    block.content = cfg.render(block.content)

                blocks.append(block)

        # regenerate preserved block content
        preserved = []
        tgt = ""
        for block in blocks:
            if block.bid is None:
                tgt += block.content
            else:
                # format cnt
                cnt = cfg.render(block.content)
                if len(cnt) == 0 or cnt[-1] != "\n":  # case where templating has eaten the remaining spaces and '\n'
                    cnt += "\n"
                preserved.append((block.bid, cnt))
                # rewrite preserved tag
                tgt += block.before_header + "{" + f"# pkglts, {block.bid}{block.after_header}\n"
                tgt += cnt
                tgt += block.before_footer + "#" + "}" + f"{block.after_footer}\n"

        ensure_path(tgt_pth)
        tgt_pth.write_text(tgt)

        return preserved



from pathlib import Path

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionTest(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            suite_name="pytest",
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        name = cfg[self._name]['suite_name']

        if name not in ("pytest", "nose"):
            invalids.append('test.suite_name')

        return invalids

    def require_option(self):
        return ['pysetup']

    def require(self, cfg):
        test_suite = cfg['test']['suite_name']
        if test_suite == 'pytest':
            yield Dependency('pytest', intent='test')
            yield Dependency('pytest-mock', intent='test')
        elif test_suite == 'nose':
            yield Dependency('nose', intent='test')
            yield Dependency('mock', intent='test')
